#!/usr/local/gauche/bin/gosh
;; -*- coding: utf-8-unix -*-

;;; one line to give the Gauche-tar and an idea of what it does.
;;; Copyright (C) 2018 aaauuue 
;;;
;;; This program is free software; you can redistribute it and/or
;;; modify it under the terms of the GNU General Public License
;;; as published by the Free Software Foundation; either version 2
;;; of the License, or (at your option) any later version.
;;;
;;; This program is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;; GNU General Public License for more details.
;;;
;;; You should have received a copy of the GNU General Public License
;;; along with this program; if not, write to the Free Software
;;; Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

(define-module tar
  (use srfi-1)
  (use srfi-35)
  (use rfc.zlib)
  (use file.util)
  (use util.match)
  (use parser.peg)
  (use gauche.uvector)
  (use gauche.vport)
  (use gauche.version)
  (use gauche.sequence)
  (use gauche.collection)
  (use gauche.mop.validator)

  (export
   <tar-entry> <tar-virtual-entry>
   <tar-reader> <debug-tar-reader> <tar-gz-reader> <debug-tar-gz-reader>
   <tar-random-reader>
   <tar-writer> <tar-builder> <tar-gz-writer> <tar-gz-builder>
   <tar-header-gnuold> <tar-header-posix> <tar-header-v7>
   <tar-error> <tar-warning> <tar-not-support>
   ;; <tar-overflow>

   tar-error? tar-error-message tar-error-debug-info
   tar-not-support? tar-not-support-message tar-not-support-debug-info
   ;; tar-overflow? tar-overflow-message tar-overflow-debug-info
   tar-warning? tar-warning-message tar-warning-debug-info

   open-output-tar-virtual-entry get-output-tar-virtual-entry
   close-output-tar-virtual-entry make-tar-virtual-entry-directory

   call-with-input-tar call-with-input-tar-from-port
   call-with-output-tar-writer call-with-output-tar-builder
   open-input-tar open-input-tar-from-port
   open-output-tar-builder open-output-tar-writer
   close-input-tar close-output-tar-builder close-output-tar-writer
   add-content tar->generator extract extract-generator

   call-with-input-tar-gz call-with-input-tar-gz-from-port
   call-with-output-tar-gz-writer call-with-output-tar-gz-builder
   open-input-tar-gz open-input-tar-gz-from-port close-input-tar-gz
   open-output-tar-gz-writer close-output-tar-gz-writer
   open-output-tar-gz-builder close-output-tar-gz-builder))

(select-module tar)


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; pax header parser
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(define %ws-set ($one-of #[ \t\r\n]))
(define %ws ($skip-many %ws-set))

;;; For Gauche 0.9.7 or later. `anychar` and `eof` of parser.peg
;;; are deprecated on Gauche 0.9.7 so I redefined it on my own.
(if (version>=? (gauche-version) "0.9.7")
  (begin
    ;;; from parser.peg.deprecated of deprecated.scm
    (define (anychar s)
      (if (pair? s)
        (return-result (car s) (cdr s))
        (return-failure/expect "character" s)))
    (define (eof s)
      (if (pair? s)
        (return-failure/expect "end of input" s)
        (return-result (eof-object) s)))))

(define %num ($lift (^x ($ x->number $ list->string x))
                    ($many ($one-of #[0-9]))))
(define %key ($lift (^x ($ string->symbol $ list->string x))
                    ($many-till anychar ($char #\=))))
(define %value ($lift (^x ($ list->string x))
                      ($many-till anychar ($or ($try %ws-set) eof))))
(define %time-format ($do (time ($sep-by %num ($char #\.))) ($return time)))
(define %record
  ($do
   (num %num)
   (_ %ws)
   (key %key)
   (_ ($char #\=))
   (value %value)
   (_ %ws)
   ($return (list num key (value-converter key value)))))
(define %records ($many %record))

(define (value-converter key value)
  (cond
   ((member key '(atime mtime ctime))
    (peg-parse-string %time-format value))
   ((member key '(size uid gid GNU.sparse.numblocks GNU.sparse.offset
                       GNU.sparse.numbytes GNU.sparse.major GNU.sparse.minor
                       GNU.sparse.size GNU.sparse.realsize GNU.volume.size
                       GNU.volume.offset))
    (string->number value))
   (else value)))

(define (parse-pax-header src)
  (peg-parse-string %records src))

(define %sep-comma-number
  ($do
   (num-list ($many-till ($sep-by %num ($char #\,))
                         ($or ($try %ws-set) eof)))
   ($return (car num-list))))

(define (parse-gnu-sparse-map src)
  (let ((lst (peg-parse-string %sep-comma-number src))
        (result '()))
    (dotimes (i (/ (length lst) 2))
      (set! result (append (list (take lst 2)) result))
      (set! lst (drop lst 2)))
    (reverse result)))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Condition
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(define-macro (define-tar-condition name super)
  `(define-condition-type ,(string->symbol (format #f "<~a>" name)) ,super
     ,(string->symbol (format #f "~a?" name))
     (debug-info ,(string->symbol (format #f "~a-debug-info" name)))
     (message ,(string->symbol (format #f "~a-message" name)))))

(define-syntax tar-condition-throw
  (syntax-rules ()
    ((_ (co r d))
     (raise
      (condition
       (co (message r) (debug-info d)))))
    ((_ ((co r d) ...))
     (raise
      (condition
       (co (message r) (debug-info d)) ...)))))

(define-tar-condition tar-error <error>)
(define-tar-condition tar-warning <error>)
;; (define-tar-condition tar-overflow <tar-error>)
(define-tar-condition tar-not-support <tar-warning>)


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Utility
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; ref: Python version 3.5 lib/tarfile.py
(define (number->base256 value :optional (digits 8))
  (let ((buf (make-uvector <u8vector> digits 0))
        (n value))
    (if (>= n 0)
      (set! #0=(uvector-ref buf 0) #o200)
      (begin
        (set! #0# #o377)
        (set! n (+ (expt 256 digits) n))))
    (dotimes (i (- digits 1))
      (set! (uvector-ref buf (- (- digits 1) i)) (logand n #o377))
      (set! n (ash n -8)))
    buf))

;;; ref: Python version 3.5 lib/tarfile.py
(define (base256->number src)
  (let* ((value 0)
         (s-list (u8vector->list src))
         (sign (car s-list)))
    (dolist (v (cdr s-list))
      (set! value (+ (ash value 8) v)))
    (cond
     ((eq? sign #o377)
      (* -1 (- (expt 256 (- (uvector-length src) 1)) value)))
     (else value))))

(define (safe-string->number src :optional (radix 8))
  (let* ((vec (string->u8vector src)))
    (if (= (uvector-length vec) 0) 0
        (cond
         ((or (eq? #0=(uvector-ref vec 0) #o200)
              (eq? #0# #o377))
          (base256->number vec))
         (else
          (if (eq? #1=(string->number src radix) #f)
            (fold + 0 (u8vector->list vec))
            #1#))))))

(define (safe-substring-u8vector src start end)
  (let ((true-end end))
    (if (> true-end #0=(string-length src))
      (set! true-end #0#))
    (string->u8vector (substring src start true-end))))

(define (tar-number->u8vector value :optional (digits 8))
  (cond
   ((<= (* -1 #0=(expt 8 digits)) value #0#)
    (string->u8vector (format #f "~o" value)))
   (else 0)
   ;; ((<= (* -1 #1=(expt 256 digits)) value #1#)
   ;;  (number->base256 value digits))
   ;; (else
   ;;  (tar-condition-throw
   ;;   (<tar-overflow>
   ;;    "Overflow. value is too big or too small."
   ;;    #"(tar-number->u8vector value :optional (digits 8)): value: [~value]")))
   ))

(define-macro (get-data cls beg end :optional (is-terminate? #t))
  `(cond
    ((eq? ,is-terminate? #t)
     (u8vector->string (uvector-copy (~ ,cls 'header) ,beg ,end)
                       0 ,(- end beg) 0))
    (else
     (u8vector->string (uvector-copy (~ ,cls 'header) ,beg ,end)
                       0 ,(- end beg)))))

(define-macro (set-data cls data beg converter)
  `(u8vector-copy! (~ ,cls 'header) ,beg (,converter ,data)))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Tar definition
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(define *tar-block-size* 512)
(define *tar-eob* (make-uvector <u8vector> *tar-block-size* 0))
(define-macro (tar-block-count size) `(div (+ ,size 511) 512))

(define-class <gnu-old-sparse> ()
  ((header :init-form #f :init-keyword :data)
   (offset :allocation :virtual
           :slot-ref (^o (safe-string->number (get-data o 0 12) 8)))
   (numbytes :allocation :virtual
             :slot-ref (^o (safe-string->number (get-data o 12 24) 8)))))

(define-class <gnu-old-sparse-header> ()
  ((header :init-form (make-uvector <u8vector> *tar-block-size* 0))
   (sparse-list :init-form (let ((sparse-list '()))
                             (dotimes (i 21)
                               (set! sparse-list
                                     (append (list (make <gnu-old-sparse>))
                                             sparse-list)))
                             (reverse sparse-list)))
   (sp :allocation :virtual
       :slot-ref (^o
                  (call/cc
                   (^(out)
                     (let ((i 0))
                       (dolist (sparse (~ o 'sparse-list))
                         (if (> i 504) (out #f))
                         (set! (~ sparse 'header)
                               (uvector-alias <u8vector> (~ o 'header)
                                              i (+ i 24)))
                         (set! i (+ i 24))))))
                  (~ o 'sparse-list)))
   (isextended :allocation :virtual
               :slot-ref (^o (safe-string->number (get-data o 504 505))))))

(define-class <tar-header-base> ()
  ((header   :init-form (make-uvector <u8vector> *tar-block-size* 0))
   (name     :allocation :virtual
             :slot-ref (^o (get-data o 0 100)) 
             :slot-set! (^(o v)
                          (set-data
                           o v 0 (cut safe-substring-u8vector <> 0 100))))
   (mode     :allocation :virtual 
             :slot-ref (^o (safe-string->number (get-data o 100 108) 10)) 
             :slot-set! (^(o v)
                          (set-data
                           o v 100 (cut tar-number->u8vector <>))))
   (uid      :allocation :virtual 
             :slot-ref (^o (safe-string->number (get-data o 108 116))) 
             :slot-set! (^(o v)
                          
                          (set-data o v 108 (cut tar-number->u8vector <>))))
   (gid      :allocation :virtual 
             :slot-ref (^o (safe-string->number (get-data o 116 124))) 
             :slot-set! (^(o v)
                          (set-data o v 116 (cut tar-number->u8vector <>))))
   (size     :allocation :virtual 
             :slot-ref (^o (safe-string->number (get-data o 124 136))) 
             :slot-set! (^(o v)
                          (set-data o v 124 (cut tar-number->u8vector <> 12))))
   (mtime    :allocation :virtual 
             :slot-ref (^o (safe-string->number (get-data o 136 148))) 
             :slot-set! (^(o v)
                          (set-data o v 136 (cut tar-number->u8vector <> 12))))
   (chksum   :allocation :virtual 
             :slot-ref (^o (safe-string->number (get-data o 148 156))) 
             :slot-set! (^(o v)
                          (set-data o v 148 (cut tar-number->u8vector <>))))
   (typeflag :allocation :virtual 
             :slot-ref (^o (get-data o 156 157 #f)) 
             :slot-set! (^(o v)
                          (set-data
                           o v 156 (cut safe-substring-u8vector <> 0 1))))
   (linkname :allocation :virtual 
             :slot-ref (^o (get-data o 157 257)) 
             :slot-set! (^(o v)
                          (set-data
                           o v 157 (cut safe-substring-u8vector <> 0 100))))))

(define-class <tar-header-v7> (<tar-header-base>)
  ((pad :allocation :virtual 
        :slot-ref (^o (get-data o 257 512)))))

(define-class <tar-header-posix> (<tar-header-base>)
  ((magic    :allocation :virtual 
             :slot-ref (^o (get-data o 257 263 #f)) 
             :slot-set! (^(o v)
                          (set-data
                           o v 257 (cut safe-substring-u8vector <> 0 6))))
   (version  :allocation :virtual 
             :slot-ref (^o (get-data o 263 265 #f)) 
             :slot-set! (^(o v)
                          (set-data
                           o v 263 (cut safe-substring-u8vector <> 0 2))))
   (uname    :allocation :virtual 
             :slot-ref (^o (get-data o 265 297)) 
             :slot-set! (^(o v)
                          (set-data
                           o v 265 (cut safe-substring-u8vector <> 0 32))))
   (gname    :allocation :virtual 
             :slot-ref (^o (get-data o 297 329)) 
             :slot-set! (^(o v)
                          (set-data
                           o v 297 (cut safe-substring-u8vector <> 0 32))))
   (devmajor :allocation :virtual 
             :slot-ref (^o (safe-string->number (get-data o 329 337))) 
             :slot-set! (^(o v)
                          (set-data o v 329 (cut tar-number->u8vector <>))))
   (devminor :allocation :virtual 
             :slot-ref (^o (safe-string->number (get-data o 337 345))) 
             :slot-set! (^(o v)
                          (set-data o v 337 (cut tar-number->u8vector <>))))
   (prefix   :allocation :virtual 
             :slot-ref (^o (get-data o 345 500)) 
             :slot-set! (^(o v)
                          (set-data
                           o v 345 (cut safe-substring-u8vector <> 0 155))))
   (pad      :allocation :virtual 
             :slot-ref (^o (get-data o 500 512)))))

(define-class <tar-header-gnuold> (<tar-header-base>)
  ((sparse-list :init-form
                (let ((sparse-list '()))
                  (dotimes (i 4)
                    (set! sparse-list
                          (append (list (make <gnu-old-sparse>)) sparse-list)))
                  (reverse sparse-list)))
   (magic       :allocation :virtual 
                :slot-ref (^o (get-data o 257 263 #f)) 
                :slot-set! (^(o v)
                             (set-data
                              o v 257 (cut safe-substring-u8vector <> 0 6))))
   (version     :allocation :virtual 
                :slot-ref (^o (get-data o 263 265 #f)) 
                :slot-set! (^(o v)
                             (set-data
                              o v 263 (cut safe-substring-u8vector <> 0 2))))
   (uname       :allocation :virtual 
                :slot-ref (^o (get-data o 265 297)) 
                :slot-set! (^(o v)
                             (set-data
                              o v 265 (cut safe-substring-u8vector <> 0 32))))
   (gname       :allocation :virtual 
                :slot-ref (^o (get-data o 297 329)) 
                :slot-set! (^(o v)
                             (set-data
                              o v 297 (cut safe-substring-u8vector <> 0 32))))
   (devmajor    :allocation :virtual 
                :slot-ref (^o (safe-string->number (get-data o 329 337))) 
                :slot-set! (^(o v)
                             (set-data o v 329 (cut tar-number->u8vector <>))))
   (devminor    :allocation :virtual 
                :slot-ref (^o (safe-string->number (get-data o 337 345))) 
                :slot-set! (^(o v)
                             (set-data o v 337 (cut tar-number->u8vector <>))))
   (atime       :allocation :virtual 
                :slot-ref (^o (safe-string->number (get-data o 345 357)))
                :slot-set! (^(o v)
                             (set-data
                              o v 345 (cut tar-number->u8vector <> 12))))
   (ctime       :allocation :virtual 
                :slot-ref (^o (safe-string->number (get-data o 357 369))) 
                :slot-set! (^(o v)
                             (set-data
                              o v 357 (cut tar-number->u8vector <> 12))))
   (offset      :allocation :virtual 
                :slot-ref (^o (get-data o 369 381)) 
                :slot-set! (^(o v)
                             (set-data
                              o v 369 (cut safe-substring-u8vector <> 0 12))))
   (longnames   :allocation :virtual 
                :slot-ref (^o (get-data o 381 385)) 
                :slot-set! (^(o v)
                             (set-data
                              o v 381 (cut safe-substring-u8vector <> 0 4))))
   (unused-pad2 :allocation :virtual 
                :slot-ref (^o (get-data o 385 386)))
   (sparse      :allocation :virtual 
                :slot-ref (^o
                           (let ((i 0))
                             (dolist (sparse-info (~ o 'sparse-list))
                               (set! (~ sparse-info 'header)
                                     (uvector-alias
                                      <u8vector> (~ o 'header) 
                                      (+ 386 (* i 24)) (+ (+ 386 (* i 24)) 24)))
                               (set! i (+ i 1))))
                           (~ o 'sparse-list))
                :slot-set! (^(o v)
                             (set-data
                              o v 386 (cut safe-substring-u8vector <> 0 96))))
   (isextended  :allocation :virtual 
                :slot-ref (^o (safe-string->number (get-data o 482 483))) 
                :slot-set! (^(o v)
                             (set-data
                              o v 482 (cut safe-substring-u8vector <> 0 1))))
   (realsize    :allocation :virtual 
                :slot-ref (^o (safe-string->number (get-data o 483 495))) 
                :slot-set! (^(o v)
                             (set-data
                              o v 483 (cut safe-substring-u8vector <> 0 12))))
   (pad         :allocation :virtual 
                :slot-ref (^o (get-data o 495 512)))))

;;; ref: golang tar/format.go
(define-class <posix-default-keywords> ()
  ((path     :init-form #f)
   (linkpath :init-form #f)
   (size     :init-form #f)
   (uid      :init-form #f)
   (gid      :init-form #f)
   (uname    :init-form #f)
   (gname    :init-form #f)
   (mtime    :init-form #f)
   (atime    :init-form #f)
   (ctime    :init-form #f)
   (charset  :init-form #f)
   (comment  :init-form #f)))

;;; ref: golang tar/format.go
(define-class <posix-extra-keywords> ()
  ((RHT.security.selinux :init-form #f)
   (SCHILY.acl.access    :init-form #f)
   (SCHILY.acl.default   :init-form #f)))

;;; ref: golang tar/format.go
(define-class <posix-gnu-keywords> ()
  ((GNU.sparse.numblocks :init-form #f)
   (GNU.sparse.offset    :init-form '())
   (GNU.sparse.numbytes  :init-form '())
   (GNU.sparse.map       :init-form #f)
   (GNU.sparse.name      :init-form #f)
   (GNU.sparse.major     :init-form #f)
   (GNU.sparse.minor     :init-form #f)
   (GNU.sparse.size      :init-form #f)
   (GNU.sparse.realsize  :init-form #f)
   (GNU.dumpdir          :init-form #f)
   (GNU.volume.label     :init-form #f)
   (GNU.volume.filename  :init-form #f)
   (GNU.volume.size      :init-form #f)
   (GNU.volume.offset    :init-form #f)))

(define-class <tar-entry> ()
  ((name                   :init-form #f)
   (mode                   :init-form #f)
   (uid                    :init-form #f)
   (gid                    :init-form #f)
   (size                   :init-form #f)
   (mtime                  :init-form #f)
   (typeflag               :init-form #f)
   (linkname               :init-form #f)
   (devmajor               :init-form #f)
   (devminor               :init-form #f)
   (atime                  :init-form #f)
   (ctime                  :init-form #f)
   (realsize               :init-form #f)
   (longnames              :init-form #f) ;;; for gnu old
   (sparse-info            :init-form ())
   (tarobj                 :init-form #f)
   (type                   :allocation :virtual
                           :init-form 'regular
                           :slot-ref (^o (typeflag->symbol (~ o 'typeflag))))
   (sparse-byte-offset     :init-form 0)  ;;; for GNU sparse format version 1.0
   (random-access-offset   :init-form #f) ;;; for random access
   (posix-xattr-list       :init-form ())
   (posix-default-keywords :init-form (make <posix-default-keywords>))
   (posix-extra-keywords   :init-form (make <posix-extra-keywords>))
   (posix-gnu-keywords     :init-form (make <posix-gnu-keywords>))))

(define-class <tar> ()
  ((name           :init-form "")
   (ioport         :init-form #f)
   (direction      :allocation :virtual
                   :slot-ref (^o
                              (cond
                               ((eq? (input-port? (~ o 'ioport)) #t)
                                'input)
                               ((eq? (output-port? (~ o 'ioport)) #t)
                                'output)
                               (else 'unknown))))
   (is-closed?     :allocation :virtual
                   :slot-ref (^o (if (~ o 'ioport)
                                   (port-closed? (~ o 'ioport)))))
   (in-generation? :init-form #f)))

(define-class <tar-gz-base> ()
  ((gzip-dorain-port :init-form #f)
   (compression-level :init-form 6
                      :validator (^(o v)
                                   (if (and (integer? v) (<= 1 v 9))
                                     v
                                     (tar-condition-throw
                                      (<tar-error>
                                       "Invalid compression level for gzip"
                                       #"==> ~v"))))))
  :metaclass <validator-meta>)

(define-class <tar-reader-base> (<tar> <collection>)
  ((skip-data? :init-form #f)
   (post-processes :init-form '())))
(define-class <tar-reader-options-mixin> ()
  ((fast-skip? :init-form #f)))
(define-class <tar-random-reader-base> () ())
(define-class <debug-tar-reader-base> (<tar-reader-base>) ())
(define-class <tar-gz-reader-base> (<tar-gz-base>) ())

(define-class <tar-reader> (<tar-reader-base> <tar-reader-options-mixin>) ())
(define-class <tar-random-reader> (<tar-reader-base> <tar-random-reader-base> <tar-reader-options-mixin>) ())
(define-class <debug-tar-reader> (<debug-tar-reader-base> <tar-reader-options-mixin>) ())

(define-class <tar-gz-reader> (<tar-gz-reader-base> <tar-reader-base>) ())
(define-class <debug-tar-gz-reader> (<tar-gz-reader-base> <debug-tar-reader-base>) ())

(define-class <tar-writer-base> (<tar>)
  ((hardlink-table :init-form (make-hash-table))))

(define-class <tar-writer> (<tar-writer-base>)
  ((filelist :init-form '())))

(define-class <tar-builder-base> (<tar-writer-base>) ())
(define-class <tar-builder> (<tar-builder-base>) ())

(define-class <tar-gz-writer> (<tar-gz-base> <tar-writer>) ())
(define-class <tar-gz-builder> (<tar-gz-base> <tar-builder-base>) ())


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Tar utility
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(define (calc-chksum header)
  (let ((total 0))
    (dotimes (i *tar-block-size*)
      (cond
       ((and (>= i 148) (< i 156))
        (set! total (+ total 32))) ;;; space
       (else
        (set! total (+ total (uvector-ref (~ header 'header) i))))))
    total))

(define-method tar-type ((header <tar-header-base>))
  (cond
   ((and (= (uvector-ref #0=(~ header 'header) 257) 0)
         (= (uvector-ref #0# 263) 0))
    'tar-v7)
   ((and (string=? (u8vector->string #0# 257 263) "ustar ")
         (string=? (u8vector->string #0# 263 265) " \0"))
    (if (not (= (logand (~ header 'mode) (lognot #o7777)) 0))
      'tar-gnu-old
      'tar-gnu))
   ((and (string=? (u8vector->string #0# 257 263) "ustar\0")
         (string=? (u8vector->string #0# 263 265) "00"))
    'tar-posix)
   (else 'tar-unknown)))

(define-method set-tar-type ((header <tar-header-base>) type)
  (cond
   ((eq? type 'tar-v7)
    (set! (uvector-ref (~ header 'header) 257) 0)
    (set! (uvector-ref (~ header 'header) 263) 0))
   ((eq? type 'tar-posix)
    (set! (~ header 'magic) "ustar\0")
    (set! (~ header 'version) "00"))
   ((eq? type 'tar-gnu)
    (set! (~ header 'magic) "ustar ")
    (set! (~ header 'version) " \0"))
   ((eq? type 'tar-gnu-old)
    (set! (~ header 'magic) "ustar\0")
    (set! (~ header 'version) "00")
    ;;; I want to set 6th digit of mode value to 1, but logior acts odd.
    ;; (let ((h (make <tar-header-gnuold>)))
    ;;   (set! (~ h 'mode) #o644)
    ;;   (set-tar-type h 'tar-gnu-old)
    ;;   (describe h))

    ;; (format "~o" (logior 420 100000)) ;;; must be 100644 (33188 decimal).
    )))

(define-method clear-tar-header ((header <tar-header-base>))
  (uvector-copy! (~ header 'header) 0 *tar-eob*))

(define-method end-of-tar? ((tar <tar>) (header <tar-header-base>))
  (if (not (= (uvector-ref #0=(~ header 'header) 0) 0))
    #f
    (begin
      (if (and (= (uvector-ref #0# 0) 0)
               (= (u8vector-dot #0# *tar-eob*) 0))
        (let ((next-header #f))
          (skip-tar-data tar header)
          (set! next-header (read-tar-header tar))
          (if (and (= (uvector-ref #1=(~ next-header 'header) 0) 0)
                   (= (u8vector-dot #1# *tar-eob*) 0))
            #t))))))

;; (define-method skip-tar-data ((tar <tar-reader-base>) (header <tar-header-base>))
;;   (let ((block-count (tar-block-count (~ header 'size))))
;;     (port-seek (~ tar 'ioport) (* block-count *tar-block-size*) SEEK_CUR)
;;     (set! (~ tar 'skip-data?) #f)))

(define-method skip-tar-data ((tar <tar-reader-base>) (header <tar-header-base>))
  (if (eq? (~ tar 'fast-skip?) #t)
    (let ((block-count (tar-block-count (~ header 'size))))
      (port-seek (~ tar 'ioport) (* block-count *tar-block-size*) SEEK_CUR))
    (let ((byte-size (* *tar-block-size* (tar-block-count (~ header 'size)))))
      (dotimes (i (div byte-size *tar-block-size*))
        (read-uvector <u8vector> *tar-block-size* (~ tar 'ioport)))))
  (set! (~ tar 'skip-data?) #f))

(define-method skip-tar-data ((tar <tar-gz-reader-base>) (header <tar-header-base>))
  (let ((byte-size (* *tar-block-size* (tar-block-count (~ header 'size)))))
    (dotimes (i (div byte-size *tar-block-size*))
      (read-uvector <u8vector> *tar-block-size* (~ tar 'ioport)))
    (set! (~ tar 'skip-data?) #f)))

(define-method read-tar-header ((tar <tar>) :optional (read-tar-type #f))
  (let ((header (make <tar-header-posix>))
        (type (or read-tar-type 'tar-unknown)))
    (set! (~ tar 'skip-data?) #t)
    (set! (~ header 'header)
          (read-uvector <u8vector> *tar-block-size* (~ tar 'ioport)))
    (set! type (tar-type header))
    (cond
     ((eq? type 'tar-gnu-old)
      (let ((new-header (make <tar-header-gnuold>)))
        (uvector-copy! (~ new-header 'header) 0 (~ header 'header))
        new-header))
     ((eq? type 'tar-v7)
      (let ((new-header (make <tar-header-v7>)))
        (uvector-copy! (~ new-header 'header) 0 (~ header 'header))
        new-header))
     (else header))))

(define-method get-tar-long-name ((tar <tar>) (header <tar-header-base>))
  (let ((block-count (tar-block-count (~ header 'size)))
        (vec #f))
    (set! vec (read-uvector <u8vector> #0=(* block-count *tar-block-size*)
                            (~ tar 'ioport)))
    (begin0 
        (u8vector->string vec 0 #0# 0)
      (set! (~ tar 'skip-data?) #f))))

(define-method extract-tar-data-to-buffer ((tar <tar>) (header <tar-header-base>) output)
  (let* ((block-count (tar-block-count #0=(~ header 'size)))
         (remain 0)
         (buf #f)
         (out output))
    (if (not (= #0# 0))
      (begin
        (set! remain (- 512 (- (* block-count *tar-block-size*) #0#)))
        (unwind-protect
         (begin 
           (dotimes (i (- block-count 1))
             (begin
               (set! buf (read-uvector <u8vector> *tar-block-size*
                                       (~ tar 'ioport)))
               (write-uvector buf out 0 *tar-block-size*)
               (uvector-copy! buf 0 *tar-eob*)))
           (begin
             (if (not (eq? buf #f)) (uvector-copy! buf 0 *tar-eob*))
             (set! buf (read-uvector <u8vector> *tar-block-size*
                                     (~ tar 'ioport)))
             (write-uvector buf out 0 remain))))))
    (set! (~ tar 'skip-data?) #f)))

(define-method write-object ((obj <tar-entry>) port)
  (format port "\n<tar-entry>\n")
  (dolist (slot (map slot-definition-name (class-slots (class-of obj))))
    (cond
     ((member slot '(posix-gnu-keywords posix-extra-keywords
                                        posix-default-keywords))
      (write-object (~ obj slot) port))
     (else (format port "~15a: ~S\n" slot (~ obj slot))))))

(define-method write-object ((obj <posix-default-keywords>) port)
  (format port "\n<posix-default-keywords>\n")
  (dolist (slot (map slot-definition-name (class-slots (class-of obj))))
    (format port "~15a: ~S\n" slot (~ obj slot))))

(define-method write-object ((obj <posix-extra-keywords>) port)
  (format port "\n<posix-extra-keywords>\n")
  (dolist (slot (map slot-definition-name (class-slots (class-of obj))))
    (format port "~15a: ~S\n" slot (~ obj slot))))

(define-method write-object ((obj <posix-gnu-keywords>) port)
  (format port "\n<posix-gnu-keywords>\n")
  (dolist (slot (map slot-definition-name (class-slots (class-of obj))))
    (format port "~15a: ~S\n" slot (~ obj slot))))

(define (copy-class-slot-value src dest)
  (dolist (slot (map slot-definition-name (class-slots (class-of src))))
    (if (class-slot-definition (class-of dest) slot)
      (if (eq? (~ dest slot) #f)
        (set! (~ dest slot) (~ src slot))))))

(define (typeflag->symbol typeflag)
  (cond
   ((or (string=? typeflag "0") (string=? typeflag "\0")
        (string=? typeflag "S"))                          'regular)
   ((string=? typeflag "1")                               'hardlink)
   ((string=? typeflag "2")                               'symlink)
   ((string=? typeflag "3")                               'character)
   ((string=? typeflag "4")                               'block)
   ((or (string=? typeflag "5") (string=? typeflag "D"))  'directory)
   ((string=? typeflag "6")                               'fifo)
   (else                                                  'unknown)))

(load "tarvirtual.scm")

;;; load function which read/write tar or tar+gz
(load "tario.scm")
