
#include "mytar.h"
#include "sparse_list.h"

void read_header( union _tar_header *header, FILE *fp )
{
     fread( ( void * )header, BLOCKSIZE, 1, fp );
}

char *get_longname( union _tar_header *header, FILE *fp )
{
     size_t  realsize    = OCT2DEC( header->v7_header.size );
     size_t  block_count = TAR_BLOCK_COUNT( realsize );
     char   *buf         = NULL;
     char   *longname    = NULL;

     buf = ( char * )malloc( sizeof( char ) * ( block_count * BLOCKSIZE ) );
     memset( buf, 0, ( block_count * BLOCKSIZE ) );
     longname = ( char * )malloc( sizeof( char ) * realsize );
     memset( longname, 0, realsize );

     fread( ( void * )buf, 1, ( block_count * BLOCKSIZE ), fp );
     strncpy( longname, buf, realsize );

     free( buf );
     return longname;
}

void dump_data( union _tar_header *header, FILE *fp )
{
     char   *name        = header->v7_header.name;
     size_t realsize     = OCT2DEC( header->v7_header.size );
     size_t  block_count = TAR_BLOCK_COUNT( realsize );
     size_t  i           = 0;
     FILE   *fout        = NULL;
     int     remain      = 0;
     char buf[BLOCKSIZE] = { '\0' };

     if( realsize == 0 ) {
          return;
     }

     remain = 512 - ( block_count * BLOCKSIZE - realsize );

     for( i = 0; i < block_count - 1; i++ ) {
          fread( ( void * )buf, 1, BLOCKSIZE, fp );
          fwrite( ( void * )buf, 1, BLOCKSIZE, stdout );
     }

     fread( ( void * )buf, 1, BLOCKSIZE, fp );
     fwrite( ( void * )buf, 1, remain, stdout );

}

void build_oldgnu_sparse_info( union _tar_header *header, FILE *fp )
{
     size_t                lastpos     = 0;
     size_t                realsize    = OCT2DEC( header->oldgnu_header.realsize );
     size_t                bcount      = TAR_BLOCK_COUNT( realsize );
     size_t                offset      = 0;
     size_t                numbytes    = 0;
     struct sparse_header  next_sparse = { '\0' };
     struct _sparse_list  *cur         = NULL;
     
     for( int i = 0; i < 4; ++i ) {
          offset = OCT2DEC( header->oldgnu_header.sp[i].offset );
          numbytes = OCT2DEC( header->oldgnu_header.sp[i].numbytes );
          if( offset == 0 && numbytes == 0 ) {
               break;
          }
          append_sparse_list( lastpos, offset, 1 );
          append_sparse_list( offset, offset + numbytes, 0 );
          lastpos = offset + numbytes;
     }

     if( header->oldgnu_header.isextended != 0 ) {
          for( ;; ) {
               fread( ( void * )&next_sparse, 1, 512, fp );
               for( int i = 0; i < 14; ++i ) {
                    /* 
                     * printf( "\t\toffset=[%lld]\n", OCT2DEC( next_sparse.sp[i].offset ) );
                     * printf( "\t\tnumbytes=[%lld]\n", OCT2DEC( next_sparse.sp[i].numbytes ) );
                     */
                    offset = OCT2DEC( next_sparse.sp[i].offset );
                    numbytes = OCT2DEC( next_sparse.sp[i].numbytes );
                    if( offset == 0 && numbytes == 0 ) {
                         break;
                    }
                    append_sparse_list( lastpos, offset, 1 );
                    append_sparse_list( offset, offset + numbytes, 0 );
                    lastpos = offset + numbytes;
               }
               /* printf( "\t\tisextended=[%x]\n", next_sparse.isextended ); */
               if( next_sparse.isextended == 0 ) {
                    break;
               }
          }
     }

     if( lastpos < realsize ) {
          append_sparse_list( lastpos, realsize, 0 );
     }

     print_sparse_list();
}

void extract_sparse_file( union _tar_header *header, FILE *fp )
{
     size_t               realsize = OCT2DEC( header->oldgnu_header.realsize );
     size_t               bcount   = TAR_BLOCK_COUNT( realsize );
     struct _sparse_list *cur      = NULL;
     FILE                *out      = NULL;
     char buf[BLOCKSIZE]           = { '\0' };
     int                  count    = 0;
     int                  remain   = 0;

     build_oldgnu_sparse_info( header, fp );

     out = fopen( header->oldgnu_header.name, "wb" );
     if( !out ) {
          fprintf( stderr, "Can't open %s.\n", header->oldgnu_header.name );
          exit( 1 );
     }

     for( cur = root; cur != NULL; cur = cur->next ) {
          if( cur->hole == 0 ) {
               /* printf( "[F]size=[%ld]\n", cur->size ); */
               count = cur->size / BLOCKSIZE;
               remain = cur->size % BLOCKSIZE;

               for( int i = 0; i < count; ++i ) {
                    fread( ( void * )buf, 1, BLOCKSIZE, fp );
                    fwrite( ( void * )buf, 1, BLOCKSIZE, out );
                    memset( buf, 0, BLOCKSIZE );
               }
               fread( ( void * )buf, 1, remain, fp );
               fwrite( ( void * )buf, 1, remain, out );
          } else {
               count = cur->size / BLOCKSIZE;
               remain = cur->size % BLOCKSIZE;

               /* printf( "[ ]size=[%ld]\n", cur->size ); */
               memset( buf, 0, BLOCKSIZE );
               for( int i = 0; i < count; ++i ) {
                    fwrite( ( void * )buf, 1, BLOCKSIZE, out );
                    memset( buf, 0, BLOCKSIZE );
               }
               fwrite( ( void * )buf, 1, remain, out );
          }
     }
     fclose( out );

     if( realsize != bcount * BLOCKSIZE ) {
          fseek( fp, ( bcount * BLOCKSIZE ) - realsize, SEEK_CUR );
     }

     free_sparse_list();
} 

void skip_gnuold_sparse_file( union _tar_header *header, FILE *fp )
{
     size_t               archive_size = OCT2DEC( header->oldgnu_header.size );
     size_t               bcount       = TAR_BLOCK_COUNT( archive_size );
     struct sparse_header next_sparse  = { '\0' };

     if( header->oldgnu_header.isextended == 1 ) {
          for( ;; ) {
               fread( ( void * )&next_sparse, 1, BLOCKSIZE, fp );
               if( next_sparse.isextended == 0 ) {
                    break;
               }
          }
     }

     fseek( fp, bcount * BLOCKSIZE, SEEK_CUR );
}

#define FREE_NAME(SRC) {if((SRC)){free((SRC)); (SRC)=NULL;}}

int main( int argc, char **argv )
{
     union _tar_header  header;
     FILE              *fp           = NULL;
     char               typeflag     = '\0';
     char              *longname     = NULL;
     char              *longlinkname = NULL;
     enum _tar_type     type;

     fp = fopen( argv[1], "rb" );
     if( !fp ) {
          fprintf( stderr, "Can't open %s.\n", argv[1] );
          exit( 1 );
     }

     for( ;; ) {
          clear_header( &header );
          read_header( &header, fp );
          if( end_of_tar( &header, fp ) == 1 ) {
               printf( "End of tar.\n" );
               break;
          }
          dump_header( &header );
          printf( "chksum = [%d]\n", calc_chksum( &header ) );
          type = get_type( &header );
          tar_type2str( type );
          /* skip_data( &header, fp ); */

          typeflag = header.v7_header.typeflag;
          switch( ( int )typeflag ) {
          case '0':
          case '\0':
               printf( "type = regular file.\n" );
               FREE_NAME( longname );
               FREE_NAME( longlinkname );
               /* 
                * {
                *      char buf[BUFSIZ];
                *      for( int i = 0; i < 5662; ++i ) {
                *           fread( ( void * )buf, 1, 512, fp );
                *           printf( "=-=%s=-=\n", buf );
                *           memset( buf, 0, 512 );
                *      }
                *      exit( 1 );
                * }
                */
               skip_data( &header, fp );
               break;

          case '1':
               printf( "type = hardlink file.\n" );
               FREE_NAME( longname );
               FREE_NAME( longlinkname );
               skip_data( &header, fp );
               break;

          case '2':
               printf( "type = symbolic link file.\n" );
               FREE_NAME( longname );
               FREE_NAME( longlinkname );
               skip_data( &header, fp );
               break;
          
          case '3':
               printf( "type = character device file.\n" );
               FREE_NAME( longname );
               FREE_NAME( longlinkname );
               skip_data( &header, fp );
               break;
          
          case '4':
               printf( "type = block device file.\n" );
               FREE_NAME( longname );
               FREE_NAME( longlinkname );
               skip_data( &header, fp );
               break;
          
          case '5':
          case 'D':
               printf( "type = directory.\n" );
               skip_data( &header, fp );
               FREE_NAME( longname );
               FREE_NAME( longlinkname );
               break;
          
          case '6':
               printf( "type = FIFO file.\n" );
               FREE_NAME( longname );
               FREE_NAME( longlinkname );
               skip_data( &header, fp );
               break;
          
          case 'L':
               printf( "type = long name.\n" );
               longname = get_longname( &header, fp );
               printf( "longname = [%s].\n", longname );
               break;

          case 'K':
               printf( "type = long link name.\n" );
               longlinkname = get_longname( &header, fp );
               printf( "longlinkname = [%s].\n", longlinkname );
               break;

          case 'x':
               printf( "type = pax extend header.\n" );
               dump_data( &header, fp );
               break;

          case 'g':
               printf( "type = pax global extend header.\n" );
               dump_data( &header, fp );
               break;

          case 'S':
               printf( "type = GNU tar sparse file.\n" );
               printf( "sparse file real size = [%ld]\n", OCT2DEC( header.oldgnu_header.realsize ) );
               /* extract_sparse_file( &header, fp ); */
               skip_gnuold_sparse_file( &header, fp );
               break;
               
          
          default:
               printf( "Not support yet or unknown type.\n" );
               skip_data( &header, fp );
               break;
          }

          printf( "\n" );
     }

     fclose( fp );
     
     return 0;
}


/* 
 * Local variables:
 * compile-command: "gcc -std=c99 -g3 -o dumptar mytar.c sparse_list.c dumptar.c"
 * End:
 */
